#include <nds.h>
#include <stdlib.h>

#define SOUND_DUTY(n)       ((n) << 24)

//---------------------------------------------------------------------------------
s32 getFreeSoundChannel() {
	u32 i;
	for (i=1; i<15; i++) {
		if ( (SCHANNEL_CR(i) & SCHANNEL_ENABLE) == 0 ) return i;
	}
	return 16;
}
 
//---------------------------------------------------------------------------------
void stopSound(void) {
  u8 ucBcl;
  for (ucBcl=0; ucBcl<16; ucBcl++)  {
    SCHANNEL_CR(ucBcl)     = 0;
  }
}

//---------------------------------------------------------------------------------
void modifySound(int sampleRate, u8 channel, u8 vol,  u8 pan, u8 format) {
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_REPEAT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
  //SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
}

//---------------------------------------------------------------------------------
void startSound(int sampleRate, const void* data, u32 bytes, u8 channel, u8 vol,  u8 pan) {
  SCHANNEL_CR(channel)     =  0;
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_SOURCE(channel) = (u32)data;
  SCHANNEL_LENGTH(channel) = bytes >> 2 ;
  SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | SOUND_8BIT;
}

//---------------------------------------------------------------------------------
void stopSoundUnique(u8 channel) {
  SCHANNEL_CR(channel)     = 0;
}

bool isFreeSoundChannel(int i) {
  if ( (SCHANNEL_CR(i) & SCHANNEL_ENABLE) == 0 ) return true;
  return false;
}

int vcount;
touchPosition first,tempPos;

//---------------------------------------------------------------------------------
void VcountHandler() {
  static int lastbut = -1;
  
  uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0;
  
  but = REG_KEYXY;
  
  if (!( (but ^ lastbut) & (1<<6))) {
  
    tempPos = touchReadXY();
  
    x = tempPos.x;
    y = tempPos.y;
    xpx = tempPos.px;
    ypx = tempPos.py;
    z1 = tempPos.z1;
    z2 = tempPos.z2;
    
  } else {
    lastbut = but;
    but |= (1 <<6);
  }
  
  if ( vcount == 80 ) {
    first = tempPos;
  } else {
    if (	abs( xpx - first.px) > 10 || abs( ypx - first.py) > 10 ||
        (but & ( 1<<6)) ) {
  
      but |= (1 <<6);
      lastbut = but;
  
    } else { 	
      IPC->mailBusy = 1;
      IPC->touchX			= x;
      IPC->touchY			= y;
      IPC->touchXpx		= xpx;
      IPC->touchYpx		= ypx;
      IPC->touchZ1		= z1;
      IPC->touchZ2		= z2;
      IPC->mailBusy = 0;
    }
  }
  IPC->buttons		= but;
  vcount ^= (80 ^ 130);
  SetYtrigger(vcount);
}

//---------------------------------------------------------------------------------
void VblankHandler(void) {
  u32 i;

  // sound code  :)
  TransferSound *snd = IPC->soundData;
  if (snd) {
    for (i=0; i<snd->count; i++) {
      u8 chan = snd->data[i].pan; // getFreeSoundChannel();
      switch (chan) {
        case 98:
          stopSoundUnique(snd->data[i].vol);
          break;
        case 99:
          stopSound();
          break;
        default:
          startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan, snd->data[i].vol, 0);
          startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan+1, snd->data[i].vol, 127);
          break;
      }
    }
  }
  IPC->soundData = 0;
} 

//---------------------------------------------------------------------------------
int main(int argc, char ** argv) {
  // Reset the clock if needed
  rtcReset();
  
  //enable sound
  powerON(POWER_SOUND);
  SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);
  IPC->soundData = 0;

  irqInit();
  irqSet(IRQ_VBLANK, VblankHandler);
  SetYtrigger(80);
  vcount = 80;
  irqSet(IRQ_VCOUNT, VcountHandler);
  irqEnable(IRQ_VBLANK | IRQ_VCOUNT);

  while (1) {
    swiWaitForVBlank();
  }
}
