/* GAME.C
  Main and miscellaneous functions for in-game, contains the main
  game loop, etc.
*/

#include "keen.h"
#include "demobox.h"

unsigned char otherplayer;
// TODO:

// seperate status boxes for the different players
// sound keen makes when he's walking into a wall
// Your Ship Needs These Parts in multiplayer

int playerbaseframe = 0;

int max_scroll_x, max_scroll_y;
unsigned char debugmode=0,acceleratemode=0;

// and this is where the magic happens
void gameloop(void)
{
int i = 0;
int enter,lastenterstate = 0,lastquit;

  if (player.x==0 || player.y==0)
  {
    crashflag=1;
    crashflag2 = levelcontrol.curlevel;
    crashflag3 = levelcontrol.episode;
    why_term_ptr = "No player start position! (flag2=levelcontrol.curlevel, flag3=levelcontrol.episode)";
  }

  if (!loadinggame)
  {
     gameloop_initialize();
  }
  else
  {
     loadinggame = 0;
     fade.mode = FADE_GO;
     fade.dir = FADE_IN;
     fade.curamt = 0;
     fade.fadetimer = 0;
     fade.rate = FADE_NORM;
  }

  // fire all guns immediately first time around
  gunfiretimer = (gunfirefreq+1);

  // if this is Mortimer's Castle, fade in and do the conversation
  // with Mortimer.
  if (levelcontrol.episode==3 && levelcontrol.curlevel==16)
  {
     if (player.isPlaying) gamepdo_SelectFrame(i);

	 gamedo_RenderScreen();
     do
     {
        gamedo_fades();
        //gamedo_AnimatedTiles();
        //SpeedThrottle();
     } while(fade.mode!=FADE_COMPLETE && !immediate_keytable[KQUIT]);

     eseq3_Mortimer();
  }

  lastquit = 1;
  do
  {
     gamedo_fades();

     // periodically make all enemy gun fixtures fire (in ep3)
     // (also ice cannons in ep1) we do this in a global variable
     // so they're all in sync. when gunfiretimer==0 all gun SE
     // objects will fire.
     if (gunfiretimer > gunfirefreq)
     {
       gunfiretimer = 0;
     }
     else gunfiretimer++;

     // gather input and copy to player.keytable[] structures
     gamedo_getInput();

     // run the player "AI" for each player in the game
     if (!map.isworldmap)
     {
       if (player.isPlaying) gamepdo_HandlePlayer(i);
     }
     else
     {
       if (player.isPlaying) gamepdo_wm_HandlePlayer(i);
     }

     gamedo_AnimatedTiles();
     gamedo_enemyai();

     gamedo_HandleFKeys();

     /* scroll triggers */
     if (!levelcontrol.gameovermode && levelcontrol.level_done==LEVEL_NOT_DONE)
     {
        ScreenIsScrolling = 0;
        if (gamedo_ScrollTriggers()) ScreenIsScrolling = 1;
     }

     // do frameskipping, and render/blit the screen if it's time
     gamedo_frameskipping();

     // when we complete a fade out flag to exit the game loop
     if (fade.mode==FADE_COMPLETE)
     {
         if (fade.dir==FADE_OUT)
         {
            demomode = DEMO_NODEMO;
            levelcontrol.level_done = LEVEL_COMPLETE;
            levelcontrol.command = LVLC_CHANGE_LEVEL;
            if (levelcontrol.curlevel != WM_MAP_NUM)
            { // exiting a level, going back to world map
              if (levelcontrol.success==1)
              { // mark level as completed on world map                
                levelcontrol.levels_completed[levelcontrol.curlevel] = 1;
              }
              levelcontrol.chglevelto = WM_MAP_NUM;
            }
         }
         else
         {
           fade.mode = NO_FADE;
         }
     }

     // when walking through the exit door don't show keen's sprite past
     // the door frame (so it looks like he walks "through" the door)
     if (levelcontrol.level_done==LEVEL_DONE_WALK)
     {
        gamepdo_walkbehindexitdoor();
     }

     // allow enter to return to main menu
     // if we're in game over mode
     enter = (immediate_keytable[KSELECT]||immediate_keytable[KJUMP]||immediate_keytable[KPOGO]);
     if (levelcontrol.gameovermode)
     {
       if (enter&&!lastenterstate)
       {
         if (fade.mode!=FADE_GO && fade.dir!=FADE_OUT)
         {
           fade.dir = FADE_OUT;
           fade.curamt = PAL_FADE_SHADES;
           fade.fadetimer = 0;
           fade.rate = FADE_NORM;
           fade.mode = FADE_GO;
         }
       }

       if (fade.mode==FADE_COMPLETE && fade.dir==FADE_OUT)
       {
         levelcontrol.command = LVLC_GAME_OVER;
       }
     }
     lastenterstate = enter;

     #ifdef NETWORK_PLAY
//       if (numplayers>1) net_senddata();
     #endif

     if (immediate_keytable[KQUIT] && !last_immediate_keytable[KQUIT])
     {
       VerifyQuit();
     }
     if (QuitState != NO_QUIT) return;

     memcpy(last_immediate_keytable, immediate_keytable, KEYTABLE_SIZE);

     // limit frame rate
     if (!acceleratemode)
     {
        SpeedThrottle();
        #ifdef NETWORK_PLAY
//          if (numplayers>1) net_sync();
        #endif
     }
     else poll_events();

  } while(!crashflag && levelcontrol.command==LVLC_NOCOMMAND);

}

// gives keycard for door doortile to player
void give_keycard(int doortile)
{
   sound_play(SOUND_GET_CARD, PLAY_NOW);
   if (doortile==DOOR_YELLOW) player.inventory.HasCardYellow = 1;
   else if (doortile==DOOR_RED) player.inventory.HasCardRed = 1;
   else if (doortile==DOOR_GREEN) player.inventory.HasCardGreen = 1;
   else if (doortile==DOOR_BLUE) player.inventory.HasCardBlue = 1;
   else if (doortile==DOOR_YELLOW_EP3) player.inventory.HasCardYellow = 1;
   else if (doortile==DOOR_RED_EP3) player.inventory.HasCardRed = 1;
   else if (doortile==DOOR_GREEN_EP3) player.inventory.HasCardGreen = 1;
   else if (doortile==DOOR_BLUE_EP3) player.inventory.HasCardBlue = 1;
   else
   {
     crashflag = 1;
     crashflag2 = doortile;
     why_term_ptr = "give_keycard(): invalid value for doortile parameter.";
   }
}

// take away the specified keycard from player p
void take_keycard(int doortile)
{
   if (doortile==DOOR_YELLOW) player.inventory.HasCardYellow = 0;
   else if (doortile==DOOR_RED) player.inventory.HasCardRed = 0;
   else if (doortile==DOOR_GREEN) player.inventory.HasCardGreen = 0;
   else if (doortile==DOOR_BLUE) player.inventory.HasCardBlue = 0;
   else if (doortile==DOOR_YELLOW_EP3) player.inventory.HasCardYellow = 0;
   else if (doortile==DOOR_RED_EP3) player.inventory.HasCardRed = 0;
   else if (doortile==DOOR_GREEN_EP3) player.inventory.HasCardGreen = 0;
   else if (doortile==DOOR_BLUE_EP3) player.inventory.HasCardBlue = 0;
}

// unregisters all animated tiles with baseframe tile
void unregister_animtiles(int tile)
{
int i;
     for(i=0;i<MAX_ANIMTILES-1;i++)
     {
        if (animtiles[i].baseframe == tile)
        {
           animtiles[i].slotinuse = 0;
        }
     }
}

void open_door(int doortile, int doorsprite, int mpx, int mpy)
{
int o,chgtotile;

   sound_play(SOUND_DOOR_OPEN, PLAY_NOW);
   take_keycard(doortile);

   /* erase door from map */
   if (levelcontrol.episode==3)
   {
     chgtotile = map.mapdata[mpx-1][mpy];
   }
   else
   {
     chgtotile = tiles[map.mapdata[mpx][mpy]].chgtile;
   }
   map_chgtile(mpx, mpy, chgtotile);
   map_chgtile(mpx, mpy+1, chgtotile);

   // replace the door tiles with a door object, which will do the animation
   o = spawn_object(mpx<<4<<CSF,mpy<<4<<CSF,OBJ_DOOR);
   objects[o].sprite = doorsprite;      
}

// checks if score is > than "extra life at" and award 1-UPs when appropriate
void extralifeat()
{
  if (player.inventory.score > player.inventory.extralifeat)
  {
    sound_play(SOUND_EXTRA_LIFE, PLAY_NOW);
    player.inventory.lives++;
    player.inventory.extralifeat += 20000;
  }
}

// have keen pick up the goodie at screen pixel position (px, py)
void keen_get_goodie(int px, int py)
{
int mpx,mpy,t;
   mpx = px>>4;
   mpy = py>>4;
   t = map.mapdata[mpx][mpy];

   if (tiles[t].pickupable)
   {  // pick up the goodie, i.e. erase it from the map
      map_chgtile(mpx, mpy, tiles[t].chgtile);
      if (tiles[t].isAnimated) map_deanimate(mpx, mpy);
   }
   else if (tiles[t].lethal)
   {  // whoah, this "goodie" isn't so good...
      killplayer();
      return;
   }

   // do whatever the goodie is supposed to do...
   if (levelcontrol.episode==1)
   {
     procgoodie_ep1(t, mpx, mpy);
   }
   else if (levelcontrol.episode==2)
   {
     procgoodie_ep2(t, mpx, mpy);
   }
   else if (levelcontrol.episode==3)
   {
     procgoodie_ep3(t, mpx, mpy);
   }
}

void initgame(void)
{
int x,y,i;

  animtiletimer = curanimtileframe = 0;
  PlatExtending = 0;

  // reset player walk frame widths
    sprites[playerbaseframe+0].xsize = PLAYERSPRITE_WIDTH;
    sprites[playerbaseframe+1].xsize = PLAYERSPRITE_WIDTH;
    sprites[playerbaseframe+2].xsize = PLAYERSPRITE_WIDTH;
    sprites[playerbaseframe+3].xsize = PLAYERSPRITE_WIDTH;

  // set gun/ice cannon fire freq
  if (levelcontrol.episode==1)
  {
    gunfirefreq = ICECANNON_FIRE_FREQ;
  }
  else
  {
    gunfirefreq = GUN_FIRE_FREQ;
  }

  // reset the ysize attribute of all doors
  sprites[DOOR_YELLOW_SPRITE].ysize = 32;
  sprites[DOOR_RED_SPRITE].ysize = 32;
  sprites[DOOR_GREEN_SPRITE].ysize = 32;
  sprites[DOOR_BLUE_SPRITE].ysize = 32;

  levelcontrol.level_done_timer = 0;
  levelcontrol.gameovermode = 0;

  // re-goodie Yorp Statue (ep1), & Vorticon Elder switch (ep2)
  if (levelcontrol.episode==1)
  {
    tiles[YORPSTATUEHEAD].goodie = 1;
    tiles[YORPSTATUEBOTTOM].goodie = 1;
    // garg statue
    tiles[433].goodie = tiles[436].goodie = 1;
    tiles[446].goodie = tiles[447].goodie = 1;
  }
  else if (levelcontrol.episode==2)
  {
    tiles[429].standgoodie = 1;
    tiles[432].isAnimated = 1;
  }

  // all objects -> not exist
 for(i=1;i<MAX_OBJECTS-1;i++) objects[i].exists = 0;
 // clear out AnimTileInUse array
 for(y=0;y<ATILEINUSE_SIZEX-1;y++)
 {
   for(x=0;x<ATILEINUSE_SIZEY-1;x++)
   {
      AnimTileInUse[x][y] = 0;
   }
 }
 // set all animated tile slots to "not in use"
 map_unregister_all_animtiles();

// initilize default sprites for objects
  memset(objdefsprites, 0, sizeof(objdefsprites));
  if (levelcontrol.episode==1)
  {
    objdefsprites[OBJ_YORP] = OBJ_YORP_DEFSPRITE;
    objdefsprites[OBJ_GARG] = OBJ_GARG_DEFSPRITE;
    objdefsprites[OBJ_BUTLER] = OBJ_BUTLER_DEFSPRITE;
    objdefsprites[OBJ_TANK] = OBJ_TANK_DEFSPRITE;
    objdefsprites[OBJ_ICECHUNK] = OBJ_ICECHUNK_DEFSPRITE;
    objdefsprites[OBJ_ICEBIT] = OBJ_ICEBIT_DEFSPRITE;
    objdefsprites[OBJ_ROPE] = OBJ_ROPE_DEFSPRITE;

    objdefsprites[OBJ_RAY] = OBJ_RAY_DEFSPRITE_EP1;
    objdefsprites[OBJ_VORT] = OBJ_VORT_DEFSPRITE_EP1;
  }
  else if (levelcontrol.episode==2)
  {
    objdefsprites[OBJ_WALKER] = OBJ_WALKER_DEFSPRITE;
    objdefsprites[OBJ_TANKEP2] = OBJ_TANKEP2_DEFSPRITE;
    objdefsprites[OBJ_BEAR] = OBJ_BEAR_DEFSPRITE;

    objdefsprites[OBJ_RAY] = OBJ_RAY_DEFSPRITE_EP2;
    objdefsprites[OBJ_VORT] = OBJ_VORT_DEFSPRITE_EP2;
    objdefsprites[OBJ_PLATFORM] = OBJ_PLATFORM_DEFSPRITE_EP2;
    objdefsprites[OBJ_BABY] = OBJ_BABY_DEFSPRITE_EP2;
  }
  else if (levelcontrol.episode==3)
  {
    objdefsprites[OBJ_FOOB] = OBJ_FOOB_DEFSPRITE;
    objdefsprites[OBJ_NINJA] = OBJ_NINJA_DEFSPRITE;
    objdefsprites[OBJ_MOTHER] = OBJ_MOTHER_DEFSPRITE;
    objdefsprites[OBJ_MEEP] = OBJ_MEEP_DEFSPRITE;
    objdefsprites[OBJ_BALL] = OBJ_BJ_DEFSPRITE;
    objdefsprites[OBJ_JACK] = OBJ_BJ_DEFSPRITE;

    objdefsprites[OBJ_RAY] = OBJ_RAY_DEFSPRITE_EP3;
    objdefsprites[OBJ_VORT] = OBJ_VORT_DEFSPRITE_EP3;
    objdefsprites[OBJ_PLATFORM] = OBJ_PLATFORM_DEFSPRITE_EP3;
    objdefsprites[OBJ_BABY] = OBJ_BABY_DEFSPRITE_EP3;
  }

  objdefsprites[OBJ_DOOR] = DOOR_YELLOW_SPRITE;
  objdefsprites[OBJ_TELEPORTER] = OBJ_TELEPORTER_DEFSPRITE;

  objdefsprites[OBJ_SECTOREFFECTOR] = BlankSprite;

// initilize game variables
  levelcontrol.level_done = LEVEL_NOT_DONE;
  animtiletimer = curanimtileframe = 0;
  DemoObjectHandle = 0;

  player.isPlaying = 1;
  player.useObject = 1; // player objects start at 1 cause DemoObject is index 0
  player.dpadcount = player.dpadlastcount = 0;
  player.hideplayer = 0;
  player.mounted = 0;
  player.ppogostick = 0;
  player.pjumping = 0;
  player.pfalling = 0;
  player.pwalking = player.playspeed = 0;
  player.pinertia_x = player.pinertia_y = 0;
  player.playpushed_x = 0;
  player.pfiring = 0;
  player.psliding = player.psemisliding = 0;
  player.pdie = 0;
  player.pfrozentime = 0;
  player.dpadcount = player.dpadlastcount = 0;
  player.ankhtime = 0;
  player.keyprocstate = 0;         // KPROC_IDLE
  player.pjustjumped = player.pjustfell = 0;

  // initilize objects used by players.
  if (player.isPlaying)
  {
    objects[player.useObject].exists = 1;
    objects[player.useObject].onscreen = 1;
    objects[player.useObject].type = OBJ_PLAYER;
    objects[player.useObject].sprite = 0;
    objects[player.useObject].onscreen = 1;
    objects[player.useObject].AssociatedWithPlayer = 1;
    objects[player.useObject].honorPriority = 1;
  }

  memset(last_immediate_keytable, 1, KEYTABLE_SIZE);

  scroll_x = 0;
  scrollx_buf = 0;
  scrollpix = 0;
  mapx = 0;
  mapxstripepos = 0;

  scroll_y = 0;
  scrolly_buf = 0;
  scrollpixy = 0;
  mapy = 0;
  mapystripepos = 0;
}

void initgamefirsttime(void)
{
int i;

   if (loadtileattributes(levelcontrol.episode)) return;

   map.firsttime = 1;

   for(i=0;i<MAX_LEVELS;i++)
     levelcontrol.levels_completed[i] = 0;

   player.inventory.extralifeat = 20000;
   player.inventory.lives = 4;
   player.godmode = 0;

   if (levelcontrol.episode==1)
   {
     player.inventory.charges = 0;
   }
   else if (levelcontrol.episode==2)
   {
     player.inventory.charges = 3;
   }
   else
   {
     player.inventory.charges = 5;
   }
   if (demomode) player.inventory.charges = 100;

   // start with pogo stick in all episodes but 1
   if (levelcontrol.episode!=1 || demomode)
     { player.inventory.HasPogo = 1; }
   else
     { player.inventory.HasPogo = 0; }

   initsprites();

   if (demomode) srand(375);
}

char spawn_object(int x, int y, int otype)
{
int i;
 // find an unused object slot
 for(i=1;i<MAX_OBJECTS-1;i++)
 {
   if (!objects[i].exists && objects[i].type != OBJ_PLAYER)
   {
     objects[i].x = x;
     objects[i].y = y;
     objects[i].type = otype;
     objects[i].sprite = objdefsprites[otype];
     objects[i].exists = 1;
     objects[i].needinit = 1;
     objects[i].onscreen = 0;
     objects[i].hasbeenonscreen = 0;
     objects[i].canbezapped = 0;
     objects[i].zapped = 0;
     objects[i].inhibitfall = 0;
     objects[i].honorPriority = 1;
     SetAllCanSupportPlayer(i, 0);
     return i;
   }
 }
 // object could not be created
 crashflag = 1;
 crashflag2 = otype;
 crashflag3 = 0;
 why_term_ptr = "Object could not be created.";
 return 0;
}

// common enemy/object ai, such as falling, setting blocked variables,
// detecting player contact, etc.
void common_enemy_ai(int o)
{
int x,y,xa,ya,xsize,ysize;
int temp;

       if (objects[o].type==OBJ_DEMOMSG) return;

       xsize = sprites[objects[o].sprite].xsize;
       ysize = sprites[objects[o].sprite].ysize;

 // set value of blockedd--should object fall?
       temp = (objects[o].y>>CSF)+ysize;
       if ((temp>>4)<<4 != temp)
       {
          objects[o].blockedd = 0;
       }
       else
       { // on a tile boundary, test if tile under object is solid
          objects[o].blockedd = 0;
          x = (objects[o].x>>CSF);
          y = (objects[o].y>>CSF)+ysize+1;
          for(xa=0;xa<xsize-2;xa+=16)
          {
            if (tiles[getmaptileat(x+xa,y)].solidfall)
            {
              objects[o].blockedd = 1;
              goto setblockedd;
            }
          }
          if (tiles[getmaptileat(x+xsize-2, y)].solidfall)
          {
             objects[o].blockedd = 1;
          }
          setblockedd: ;
       }

  // set blockedu
    objects[o].blockedu = 0;
    x = (objects[o].x>>CSF);
    y = (objects[o].y>>CSF)-1;
    for(xa=0;xa<xsize;xa+=16)
    {
        if (tiles[getmaptileat(x+xa,y)].solidceil)
        {
          objects[o].blockedu = 1;
          goto setblockedu;
        }
    }
    if (tiles[getmaptileat(x+xsize-1, y)].solidceil)
    {
      objects[o].blockedu = 1;
    }
    setblockedu: ;

 // set blockedl
    objects[o].blockedl = 0;
    x = (objects[o].x>>CSF)-1;
    y = (objects[o].y>>CSF)+1;
    for(ya=0;ya<ysize;ya+=16)
    {
        if (tiles[getmaptileat(x,y+ya)].solidr)
        {
          objects[o].blockedl = 1;
          goto setblockedl;
        }
    }
    if (tiles[getmaptileat(x, ((objects[o].y>>CSF)+ysize-1))].solidr)
    {
      objects[o].blockedl = 1;
    }
    setblockedl: ;

 // set blockedr
    objects[o].blockedr = 0;
    x = (objects[o].x>>CSF)+xsize;
    y = (objects[o].y>>CSF)+1;
    for(ya=0;ya<ysize;ya+=16)
    {
        if (tiles[getmaptileat(x,y+ya)].solidl)
        {
          objects[o].blockedr = 1;
          goto setblockedr;
        }
    }
    if (tiles[getmaptileat(x, ((objects[o].y>>CSF)+ysize-1))].solidl)
    {
      objects[o].blockedr = 1;
    }
    setblockedr: ;
  
    // hit detection with players
    objects[o].touchPlayer = 0;
    if (player.isPlaying)
    {
      objects[player.useObject].x = player.x;
      objects[player.useObject].y = player.y;
      objects[player.useObject].sprite = 0;
      if (!player.pdie && hitdetect(o, player.useObject))
          objects[o].touchPlayer = 1;
    }

// have object fall if it should
  if (!objects[o].inhibitfall)
  {
       #define OBJFALLSPEED   20
       if (objects[o].blockedd)
       {
         objects[o].yinertia = 0;
       }
       else
       {
#define OBJ_YINERTIA_RATE  5
         if (objects[o].yinertiatimer>OBJ_YINERTIA_RATE)
         {
           if (objects[o].yinertia < OBJFALLSPEED) objects[o].yinertia++;
           objects[o].yinertiatimer = 0;
         } else objects[o].yinertiatimer++;
       }
       objects[o].y += objects[o].yinertia;
  }
}

// returns nonzero if object1 overlaps object2
char hitdetect(int object1, int object2)
{
int s1, s2;
unsigned int rect1x1, rect1y1, rect1x2, rect1y2;
unsigned int rect2x1, rect2y1, rect2x2, rect2y2;

  // get the sprites used by the two objects
  s1 = objects[object1].sprite;
  s2 = objects[object2].sprite;

  // get the bounding rectangle of the first object
  rect1x1 = objects[object1].x + sprites[s1].bboxX1;
  rect1y1 = objects[object1].y + sprites[s1].bboxY1;
  rect1x2 = objects[object1].x + sprites[s1].bboxX2;
  rect1y2 = objects[object1].y + sprites[s1].bboxY2;

  // get the bounding rectangle of the second object
  rect2x1 = objects[object2].x + sprites[s2].bboxX1;
  rect2y1 = objects[object2].y + sprites[s2].bboxY1;
  rect2x2 = objects[object2].x + sprites[s2].bboxX2;
  rect2y2 = objects[object2].y + sprites[s2].bboxY2;

  // find out if the rectangles overlap
  if ((rect1x1 < rect2x1) && (rect1x2 < rect2x1)) return 0;
  if ((rect1x1 > rect2x2) && (rect1x2 > rect2x2)) return 0;
  if ((rect1y1 < rect2y1) && (rect1y2 < rect2y1)) return 0;
  if ((rect1y1 > rect2y2) && (rect1y2 > rect2y2)) return 0;

  return 1;
}

void killplayer()
{
   if (player.godmode || immediate_keytable[KTAB]) return;
   if (player.ankhtime) return;
   if (levelcontrol.level_done) return;
   if (!player.pdie)
   {
      player.pdie = PDIE_DYING;
      player.pdieframe = 0;
      player.pdietimer = 0;
      player.pdietillfly = DIE_TILL_FLY_TIME;
      player.pdie_xvect = rand()%(DIE_MAX_XVECT*2);
      player.pdie_xvect -= DIE_MAX_XVECT;
      player.inventory.lives--;
      player.y += (8<<CSF);
      gamepdo_SelectFrame();
      sound_play(SOUND_KEEN_DIE, PLAY_NOW);
   }
}

void freezeplayer()
{
   if (player.godmode || immediate_keytable[KTAB]) return;
   if (player.ankhtime) return;
   // give the player a little "kick"
   player.pjumptime = PJUMP_NORMALTIME_1;
   player.pjumpupdecreaserate = PJUMP_UPDECREASERATE_1;
   player.pjumpupspeed = 15;
   player.pjumping = PJUMPUP;
   player.pjumpupspeed_decreasetimer = 0;
   player.pjustjumped = 1;

   // and freeze him (stun him on ep2/3)
   player.pfrozentime = PFROZEN_TIME;
   player.pfrozenframe = 0;
   player.pfrozenanimtimer = 0;
   player.ppogostick = 0;
}


void PlayerTouchedExit()
{
       if (!player.pjumping && !player.pfalling\
           && !player.ppogostick && \
           levelcontrol.level_done==LEVEL_NOT_DONE)
       {
            // don't allow player to walk through a door if he's standing
            // on an object such as a platform or an enemy
            if (player.psupportingobject)
            {
              return;
            }

            // if player has ankh shut it off
            if (player.ankhtime)
            {
              player.ankhtime = 0;
              objects[player.ankhshieldobject].exists = 0;
            }

            player.ppogostick = 0;

            sound_play(SOUND_LEVEL_DONE, PLAY_NOW);
            levelcontrol.level_done = LEVEL_DONE_WALK;
       }
}

void endlevel(int success)
{
  if (fade.mode == NO_FADE)
  {
    fade.dir = FADE_OUT;
    fade.curamt = PAL_FADE_SHADES;
    fade.fadetimer = 0;
    fade.rate = FADE_NORM;
    fade.mode = FADE_GO;
    levelcontrol.success = success;
    levelcontrol.tobonuslevel = 0;
  }
}

void SetGameOver(void)
{
   if (!levelcontrol.gameovermode)
   {
     levelcontrol.gameovermode = 1;
     sound_play(SOUND_GAME_OVER, PLAY_NOW);
   }
}


// this is so objects can block the player,
// player can stand on them, etc.
// x and y are the CSFed coordinates to check (e.g. playx and playy)
// returns nonzero if there is a solid object
// at that point
char checkobjsolid(int x, int y)
{
  int o;

   for(o=1;o<MAX_OBJECTS;o++)
   {
      if (objects[o].exists && objects[o].cansupportplayer)
      {
        if (x >= objects[o].x+sprites[objects[o].sprite].bboxX1)
          if (x <= objects[o].x+sprites[objects[o].sprite].bboxX2)
            if (y >= objects[o].y+sprites[objects[o].sprite].bboxY1)
              if (y <= objects[o].y+sprites[objects[o].sprite].bboxY2)
                return o;
      }
   }
 return 0;
}

// returns 1 if player cp has the card to door t
unsigned char CheckDoorBlock(int t)
{
   if (levelcontrol.episode!=3)
   {  // episode 1-2
        if (t==DOOR_YELLOW || t==DOOR_YELLOW+1)
        {
          if (!player.inventory.HasCardYellow)
          {
            player.blockedby = DOOR_YELLOW;
            return 1;
          }
        }
        else if (t==DOOR_RED || t==DOOR_RED+1)
        {
          if (!player.inventory.HasCardRed)
          {
            player.blockedby = DOOR_RED;
            return 1;
          }
        }
        else if (t==DOOR_GREEN || t==DOOR_GREEN+1)
        {
          if (!player.inventory.HasCardGreen)
          {
            player.blockedby = DOOR_GREEN;
            return 1;
          }
        }
        else if (t==DOOR_BLUE || t==DOOR_BLUE+1)
        {
          if (!player.inventory.HasCardBlue)
          {
            player.blockedby = DOOR_BLUE;
            return 1;
          }
        }
   }
   else
   {  // episode 3
        if (t==DOOR_YELLOW_EP3 || t==DOOR_YELLOW_EP3+1)
        {
          if (!player.inventory.HasCardYellow)
          {
            player.blockedby = DOOR_YELLOW_EP3;
            return 1;
          }
        }
        else if (t==DOOR_RED_EP3 || t==DOOR_RED_EP3+1)
        {
          if (!player.inventory.HasCardRed)
          {
            player.blockedby = DOOR_RED_EP3;
            return 1;
          }
        }
        else if (t==DOOR_GREEN_EP3 || t==DOOR_GREEN_EP3+1)
        {
          if (!player.inventory.HasCardGreen)
          {
            player.blockedby = DOOR_GREEN_EP3;
            return 1;
          }
        }
        else if (t==DOOR_BLUE_EP3 || t==DOOR_BLUE_EP3+1)
        {
          if (!player.inventory.HasCardBlue)
          {
            player.blockedby = DOOR_BLUE_EP3;
            return 1;
          }
        }
   }

   return 0;
}

// checks if tile at (x,y) is solid to the player walking left into it.
// returns 1 and sets blockedby if so.
char checkissolidl(int x, int y)
{
int t;
  t = getmaptileat(x, y);
  if (tiles[t].solidl)
  {
    player.blockedby = t;
    return 1;
  }
  if (checkobjsolid(x<<CSF,y<<CSF))
  {
    player.blockedby = 0;
    return 1;
  }
  else
  {
    // don't let player walk through doors he doesn't have the key to
    if (CheckDoorBlock(t))
    {
      return 1;
    }
  }
  return 0;
}

// checks if tile at (x,y) is solid to the player walking right into it.
// returns 1 and sets blockedby if so.
char checkissolidr(int x, int y)
{
int t;
  t = getmaptileat(x, y);
  if (tiles[t].solidr)
  {
    player.blockedby = t;
    return 1;
  }
  else if (checkobjsolid(x<<CSF,y<<CSF))
  {
    player.blockedby = 0;
    return 1;
  }
  else
  {
    // don't let player walk through doors he doesn't have the key to
    if (CheckDoorBlock(t))
    {
      return 1;
    }
  }
  return 0;
}

// copies tile data into a sprite. multiple tiles can be copied into
// a single sprite--they will be stacked up vertically.
void CopyTileToSprite(int t, int s, int ntilestocopy, int transparentcol)
{
int x,y1,y2,tboundary;
    sprites[s].xsize = 16;
    sprites[s].ysize = 16 * ntilestocopy;
    sprites[s].bboxX1 = sprites[s].bboxY1 = 0;
    sprites[s].bboxX2 = (sprites[s].xsize << CSF);
    sprites[s].bboxY2 = (sprites[s].ysize << CSF);

    tboundary = 0;
    y2 = 0;
    for(y1=0;y1<(16*ntilestocopy);y1++)
    {
      for(x=0;x<16;x++)
      {
        sprites[s].imgdata[y1][x] = tiledata[t][y2][x];
        if (sprites[s].imgdata[y1][x] != transparentcol)
        {
          sprites[s].maskdata[y1][x] = 15;
        }
        else
        {
          sprites[s].maskdata[y1][x] = 0;
        }
      }
      y2++;
      if (y2>=16)
      {
        y2 = 0;
        t++;
      }
    }
}

// creates a mask from a sourcetile, places it in desttile
void MakeMask(int sourcetile, int desttile, int transparentcol)
{
int x,y,c;
   for(y=0;y<16;y++)
   {
     for(x=0;x<16;x++)
     {
        c = tiledata[sourcetile][y][x];
        if (c != transparentcol) c = 0; else c = 15;
        tiledata[desttile][y][x] = c;
     }
   }
}

// replaces all instances of color find in sprite s with
// color replace, as long as the y is greater than miny
void ReplaceSpriteColor(int s, int find, int replace, int miny)
{
int x,y;

 for(y=miny;y<sprites[s].ysize;y++)
 {
   for(x=0;x<sprites[s].xsize;x++)
   {
     if (sprites[s].imgdata[y][x]==find)
     {
       sprites[s].imgdata[y][x] = replace;
     }
   }
 }
}

// duplicates sprite source to dest
void CopySprite(int source, int dest)
{
int x,y;

  sprites[dest].xsize = sprites[source].xsize;
  sprites[dest].ysize = sprites[source].ysize;

  for(y=0;y<sprites[source].ysize;y++)
  {
    for(x=0;x<sprites[source].xsize;x++)
    {
      sprites[dest].imgdata[y][x] = sprites[source].imgdata[y][x];
      sprites[dest].maskdata[y][x] = sprites[source].maskdata[y][x];
    }
  }
}

// initilize sprites that come from tiles, such as the doors
void initsprites(void)
{
int x,y,s,indx;

    if (levelcontrol.episode!=3)
    {
      CopyTileToSprite(DOOR_YELLOW, DOOR_YELLOW_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_RED, DOOR_RED_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_GREEN, DOOR_GREEN_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_BLUE, DOOR_BLUE_SPRITE, 2, 7);
    }
    else
    {
      CopyTileToSprite(DOOR_YELLOW_EP3, DOOR_YELLOW_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_RED_EP3, DOOR_RED_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_GREEN_EP3, DOOR_GREEN_SPRITE, 2, 7);
      CopyTileToSprite(DOOR_BLUE_EP3, DOOR_BLUE_SPRITE, 2, 7);
    }

    // create BLANKSPRITE
    s = LatchHeader.NumSprites;
    BlankSprite = s;
    sprites[BlankSprite].xsize = sprites[BlankSprite].ysize = 0;

    // create DEMOSPRITE (the "demo" message the appears on the screen
    // during a demo--the graphics for it are stored in demobox.h)
    s++;
    DemoSprite = s;
    sprites[DemoSprite].xsize = 48;
    sprites[DemoSprite].ysize = 16;
    indx = 0;
    for(y=0;y<DEMOBOX_HEIGHT;y++)
    {
      for(x=0;x<DEMOBOX_WIDTH;x++)
      {
        sprites[DemoSprite].imgdata[y][x] = demobox_image[indx];
        sprites[DemoSprite].maskdata[y][x] = demobox_mask[indx];
        indx++;
      }
    }

    // set up mask for ice cannons (ep1)
    if (levelcontrol.episode==1)
    {

//       MakeMask(443, 155, 7);
  //     tiles[443].masktile = 155;
//       tiles[428].masktile = 155;
    }
}

void procgoodie_ep1(int t, int mpx, int mpy)
{
   if (t>=201 && t<=205) sound_play(SOUND_GET_BONUS, PLAY_NOW);
   else if (t==175||t==176) sound_play(SOUND_GET_ITEM, PLAY_NOW);
   switch(t)
   {
    // keycards
    case 190: give_keycard(DOOR_YELLOW); break;
    case 191: give_keycard(DOOR_RED); break;
    case 192: give_keycard(DOOR_GREEN); break;
    case 193: give_keycard(DOOR_BLUE); break;

    case DOOR_YELLOW:
           if (player.inventory.HasCardYellow)
             open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy);
         break;
    case DOOR_YELLOW+1:
           if (player.inventory.HasCardYellow)
             open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy-1);
         break;
    case DOOR_RED:
           if (player.inventory.HasCardRed)
             open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy);
         break;
    case DOOR_RED+1:
           if (player.inventory.HasCardRed)
             open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy-1);
         break;
    case DOOR_GREEN:
           if (player.inventory.HasCardGreen)
             open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy);
         break;
    case DOOR_GREEN+1:
           if (player.inventory.HasCardGreen)
             open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy-1);
         break;
    case DOOR_BLUE:
           if (player.inventory.HasCardBlue)
             open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy);
         break;
    case DOOR_BLUE+1:
           if (player.inventory.HasCardBlue)
             open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy-1);
         break;
    
    case 202:    // lollipop
         player.inventory.score += 100; extralifeat();
         break;
    case 203:    // pepsi
         player.inventory.score += 200; extralifeat();
         break;
    case 201:    // pizza
         player.inventory.score += 500; extralifeat();
         break;
    case 204:    // book
         player.inventory.score += 1000; extralifeat();
         break;
    case 205:    // teddy bear
         player.inventory.score += 5000; extralifeat();
         break;

    case 175:           // raygun
         player.inventory.charges += 5;
    break;
    case 176:           // the Holy Pogo Stick
         player.inventory.HasPogo = 1;
    break;
    
    case 221: case 222: case 223: case 224:   //joystick
     player.inventory.HasJoystick = 1;
     sound_play(SOUND_GET_PART, PLAY_NOW);
    break;
    case 237: case 238: case 239: case 240:   //battery
     player.inventory.HasBattery = 1;
     sound_play(SOUND_GET_PART, PLAY_NOW);
    break;
    case 241: case 242: case 243: case 244:   //vacuum cleaner
     player.inventory.HasVacuum = 1;
     sound_play(SOUND_GET_PART, PLAY_NOW);
    break;
    case 245: case 246: case 247: case 248:   //fuel
     player.inventory.HasFuel = 1;
     sound_play(SOUND_GET_PART, PLAY_NOW);
    break;

    // in-level teleporter
    // (in level13.ck1 that takes you to the bonus level)
    case 481:
    case 494:
        endlevel(0);
        levelcontrol.tobonuslevel = 1;
    break;

    case YORPSTATUEHEAD:
      youseeinyourmind(mpx, mpy, 0);
      tiles[YORPSTATUEHEAD].goodie = 0;
      tiles[YORPSTATUEBOTTOM].goodie = 0;
      break;
    case YORPSTATUEBOTTOM:
      youseeinyourmind(mpx, mpy-1, 0);
      tiles[YORPSTATUEHEAD].goodie = 0;
      tiles[YORPSTATUEBOTTOM].goodie = 0;
      break;
    // garg statue top-left
    case 433:
      youseeinyourmind(mpx, mpy, 1);
      tiles[433].goodie = tiles[436].goodie = 0;
      tiles[446].goodie = tiles[447].goodie = 0;
      break;
    // garg statue top-right when it's glowing
    case 436:
      youseeinyourmind(mpx-1, mpy, 1);
      tiles[433].goodie = tiles[436].goodie = 0;
      tiles[446].goodie = tiles[447].goodie = 0;
      break;
    // garg statue bottom-left
    case 446:
      youseeinyourmind(mpx, mpy-1, 1);
      tiles[433].goodie = tiles[436].goodie = 0;
      tiles[446].goodie = tiles[447].goodie = 0;
      break;
    // garg statue bottom-right
    case 447:
      youseeinyourmind(mpx-1, mpy-1, 1);
      tiles[433].goodie = tiles[436].goodie = 0;
      tiles[446].goodie = tiles[447].goodie = 0;
      break;
    // exit door
    case 159:
      if (levelcontrol.canexit && (!levelcontrol.isfinallevel || player.inventory.HasFuel))
      {
        levelcontrol.exitXpos = (mpx+2)<<4;
        PlayerTouchedExit();
      }
    break;

    // we fell off the bottom of the map
    case TILE_FELLOFFMAP_EP1:
      if (!player.pdie)
      {
        sound_play(SOUND_KEEN_FALL, PLAY_FORCE);
        player.ankhtime = 0;
        player.godmode = 0;
        player.pdie = PDIE_FELLOFFMAP;
      }
      break;

    default: crashflag = 1;
             crashflag2 = t;
             why_term_ptr = "procgoodie_ep1: Unknown goodie-- value given in flag2.";
             break;
   }


}

void procgoodie_ep2(int t, int mpx, int mpy)
{
   switch(t)
   {
    // keycards
    case 190: give_keycard(DOOR_YELLOW); break;
    case 191: give_keycard(DOOR_RED); break;
    case 192: give_keycard(DOOR_GREEN); break;
    case 193: give_keycard(DOOR_BLUE); break;

    case DOOR_YELLOW:
           if (player.inventory.HasCardYellow)
             open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy);
         break;
    case DOOR_YELLOW+1:
           if (player.inventory.HasCardYellow)
             open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy-1);
         break;
    case DOOR_RED:
           if (player.inventory.HasCardRed)
             open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy);
         break;
    case DOOR_RED+1:
           if (player.inventory.HasCardRed)
             open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy-1);
         break;
    case DOOR_GREEN:
           if (player.inventory.HasCardGreen)
             open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy);
         break;
    case DOOR_GREEN+1:
           if (player.inventory.HasCardGreen)
             open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy-1);
         break;
    case DOOR_BLUE:
           if (player.inventory.HasCardBlue)
             open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy);
         break;
    case DOOR_BLUE+1:
           if (player.inventory.HasCardBlue)
             open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy-1);
         break;
    
    case 202:    // candy bar
    case 306:
         player.inventory.score += 100; extralifeat();
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         break;
    case 203:    // hamburger
    case 307:
         player.inventory.score += 200; extralifeat();
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         break;
    case 201:    // coke
    case 308:
         player.inventory.score += 500; extralifeat();
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         break;
    case 204:    // cake
    case 309:
         player.inventory.score += 1000; extralifeat();
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         break;
    case 205:    // bear
    case 310:
         player.inventory.score += 5000; extralifeat();
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         break;

    case 175:           // raygun
    case 311:
         player.inventory.charges += 5;
         sound_play(SOUND_GET_ITEM, PLAY_NOW);
    break;

    case 176: break;    // pogo (not used in ep2)

    case 429:           // switch for vorticon elder
      VorticonElder(mpx, mpy);
      tiles[429].standgoodie = 0;
    break;
    
    // exit door
    case 159:
      if (levelcontrol.canexit)
      {
        levelcontrol.exitXpos = (mpx+2)<<4;
        PlayerTouchedExit();
      }
    break;

    default: crashflag = 1;
             crashflag2 = t;
             why_term_ptr = "procgoodie_ep2: Unknown goodie-- value given in flag2.";
             break;
   }

}

void procgoodie_ep3(int t, int mpx, int mpy)
{
   switch(t)
   {
    case 144:    // candy bar
    case 144+13:
    case 144+13+13:
    case 144+13+13+13:
    case 144+13+13+13+13:
    case 144+13+13+13+13+13:
         player.inventory.score += 100;
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         extralifeat();
         break;
    case 145:    // hamburger
    case 145+13:
    case 145+13+13:
    case 145+13+13+13:
    case 145+13+13+13+13:
    case 145+13+13+13+13+13:
         player.inventory.score += 200;
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         extralifeat();
         break;
    case 146:    // coke
    case 146+13:
    case 146+13+13:
    case 146+13+13+13:
    case 146+13+13+13+13:
    case 146+13+13+13+13+13:
         player.inventory.score += 500;
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         extralifeat();
         break;
    case 147:    // cake
    case 147+13:
    case 147+13+13:
    case 147+13+13+13:
    case 147+13+13+13+13:
    case 147+13+13+13+13+13:
         player.inventory.score += 1000;
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         extralifeat();
         break;
    case 148:    // bear
    case 148+13:
    case 148+13+13:
    case 148+13+13+13:
    case 148+13+13+13+13:
    case 148+13+13+13+13+13:
         player.inventory.score += 5000;
         sound_play(SOUND_GET_BONUS, PLAY_NOW);
         extralifeat();
         break;
    case 149:    // ankh
    case 149+13:
    case 149+13+13:
    case 149+13+13+13:
    case 149+13+13+13+13:
    case 149+13+13+13+13+13:
         GiveAnkh();
         break;
    case 150:   // raygun charge
    case 150+13:
    case 150+13+13:
    case 150+13+13+13:
    case 150+13+13+13+13:
    case 150+13+13+13+13+13:
      player.inventory.charges++;
      sound_play(SOUND_GET_ITEM, PLAY_NOW);
      break;
    case 151:   // raygun
    case 151+13:
    case 151+13+13:
    case 151+13+13+13:
    case 151+13+13+13+13:
    case 151+13+13+13+13+13:
      player.inventory.charges += 5;
      sound_play(SOUND_GET_ITEM, PLAY_NOW);
      break;
    case 152:   // yellow keycard
    case 152+13:
    case 152+13+13:
    case 152+13+13+13:
    case 152+13+13+13+13:
    case 152+13+13+13+13+13:
      give_keycard(DOOR_YELLOW_EP3);
      break;
    case 153:   // red keycard
    case 153+13:
    case 153+13+13:
    case 153+13+13+13:
    case 153+13+13+13+13:
    case 153+13+13+13+13+13:
      give_keycard(DOOR_RED_EP3);
      break;
    case 154:   // green keycard
    case 154+13:
    case 154+13+13:
    case 154+13+13+13:
    case 154+13+13+13+13:
    case 154+13+13+13+13+13:
      give_keycard(DOOR_GREEN_EP3);
      break;
    case 155:   // blue keycard
    case 155+13:
    case 155+13+13:
    case 155+13+13+13:
    case 155+13+13+13+13:
    case 155+13+13+13+13+13:
      give_keycard(DOOR_BLUE_EP3);
      break;
    case DOOR_YELLOW_EP3:
         if (player.inventory.HasCardYellow)
           open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy);
         break;
    case DOOR_YELLOW_EP3+1:
         if (player.inventory.HasCardYellow)
           open_door(DOOR_YELLOW, DOOR_YELLOW_SPRITE, mpx, mpy-1);
         break;
    case DOOR_RED_EP3:
         if (player.inventory.HasCardRed)
           open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy);
         break;
    case DOOR_RED_EP3+1:
         if (player.inventory.HasCardRed)
           open_door(DOOR_RED, DOOR_RED_SPRITE, mpx, mpy-1);
         break;
    case DOOR_GREEN_EP3:
         if (player.inventory.HasCardGreen)
           open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy);
         break;
    case DOOR_GREEN_EP3+1:
         if (player.inventory.HasCardGreen)
           open_door(DOOR_GREEN, DOOR_GREEN_SPRITE, mpx, mpy-1);
         break;
    case DOOR_BLUE_EP3:
         if (player.inventory.HasCardBlue)
           open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy);
         break;
    case DOOR_BLUE_EP3+1:
         if (player.inventory.HasCardBlue)
           open_door(DOOR_BLUE, DOOR_BLUE_SPRITE, mpx, mpy-1);
         break;

    case 78:
    // cave spikes pointing up
      killplayer();
    break;

    // left side of exit door
    case 255:
    case 242:
      if (levelcontrol.canexit)
      {
        levelcontrol.exitXpos = (mpx+2)<<4;
        PlayerTouchedExit();
      }
      break;
    // middle of exit door (right side of green part)
    case 243:
    case 256:
    break;
    /*
      if (levelcontrol.canexit)
      {
        levelcontrol.exitXpos = (mpx+1)<<4;
        PlayerTouchedExit(theplayer);
      }
      break;*/
    // we fell off the bottom of the map
    case TILE_FELLOFFMAP_EP3:
      if (!player.pdie)
      {
        sound_play(SOUND_KEEN_FALL, PLAY_FORCE);
        player.ankhtime = 0;
        player.godmode = 0;
        player.pdie = PDIE_FELLOFFMAP;
      }
      break;

    default: crashflag = 1;
             crashflag2 = t;
             why_term_ptr = "procgoodie_ep3: Unknown goodie-- value given in flag2.";
             break;
   }
}

void GiveAnkh()
{
int o;
  if (!player.ankhtime)
  {
    o = spawn_object(player.x, player.y, OBJ_SECTOREFFECTOR);
    objects[o].ai.se.type = SE_ANKHSHIELD;
    player.ankhshieldobject = o;
  }

  sound_play(SOUND_ANKH, PLAY_NOW);
  player.ankhtime = PLAY_ANKH_TIME;
  gamepdo_ankh();
}

void gameloop_initialize(void)
{
int x,y,i;
int timeout;

   if (levelcontrol.episode == 3)
   {
     // coat the top of the map ("oh no!" border) with a non-solid tile
     // so keen can jump partially off the top of the screen
     for(x=1;x<map.xsize-2;x++)
     {
         map.mapdata[x][1] = 143;
     }

     // make it lethal to fall off the bottom of the map
     // the "oh no" border is set solidceil=0, and here we're
     // going to coat the bottom of the map, below the "oh no"
     // border, with a special tile that has standgoodie set
     // and will trigger the "fell off the map" kill
     y = map.ysize;
     for(x=2;x<map.xsize-2;x++)
     {
        map.mapdata[x][y] = TILE_FELLOFFMAP_EP3;
     }
   }

  if (levelcontrol.episode==1)
  {
    // coat the bottom of the map below the border.
    // since the border has solidceil=1 this provides
    // a platform to catch yorps that fall off the map
    y = map.ysize;
    for(x=2;x<map.xsize-2;x++)
    {
       map.mapdata[x][y] = TILE_FELLOFFMAP_EP1;
    }
  }                             

   // draw map first time
   drawmap();

   // set the maximum amount the map is allowed to scroll
   max_scroll_x = (map.xsize-2-(320/16))<<4;
   max_scroll_y = (map.ysize<<4)-200-32;

   // players start facing left if the start position
   // is on the left half of the map, else right
   // for ep1 it's always facing right (because of level08)

    if (map.isworldmap)
      player.pdir = player.pshowdir = DOWN;
    else
    {
      if (player.x>>CSF>>4 < (map.xsize/2) || levelcontrol.episode==1)
        player.pdir = player.pshowdir = RIGHT;
      else
        player.pdir = player.pshowdir = LEFT;
    }

// scroll past the first two tiles (the level border), they'll
// now never be visible because you're not allowed to scroll
// left past X=32.
   for(i=0;i<2*16;i++)
   {
      map_scroll_right();
      map_scroll_down();
   }

// scroll the screen until the primary player is onscreen
// enough to where he doesn't set off the scroll triggers
  for(timeout=0;timeout<10000;timeout++)
  {
    if (!gamedo_ScrollTriggers()) break;
  }

  // initiate fade-in
  fade.mode = FADE_GO;
  fade.dir = FADE_IN;
  fade.rate = FADE_NORM;
  fade.curamt = 0;
  fade.fadetimer = 0;

  // "keens left" when returning to world map after dying
  if (levelcontrol.dokeensleft)
  {
    keensleft();
    levelcontrol.dokeensleft = 0;
  }

}
