/* GM_PDOWM.C
  Contains all of the gamepdo_wm_xxx functions...which are called from the
  main game loop. These functions perform some task that is done each
  time around the game loop and is related to the player when he is on
  the world map. Basically gamepdo.c but has slightly different versions
  of some functions to handle the world map.
*/

#include "keen.h"

void AllowMountUnmountNessie();
void MountNessieIfAvailable();

// player handler mother-function, calls all needed "gamepdo" functions
// world map edition
void gamepdo_wm_HandlePlayer()
{
    player.inhibitwalking = 0;
    player.inhibitfall = 0;

    gamepdo_StatusBox();

    gamepdo_ProcessInput();
    gamepdo_wm_setdir();

    gamepdo_wm_setblockedlrud();
    gamepdo_wm_AllowEnterLevel();

    gamepdo_walking();
    gamepdo_walkinganim();

    if (fade.mode==NO_FADE)
    {
      gamepdo_InertiaAndFriction_X();
      gamepdo_InertiaAndFriction_Y();
    }

    if (levelcontrol.episode==3)
    {
      AllowMountUnmountNessie();
    }

    gamepdo_wm_SelectFrame();

    // copy player's keytable to lastkeytable
    memcpy(&player.lastkeytable, &player.keytable, sizeof(player.lastkeytable));
}

// select the appropriate player frame based on what he's doing
void gamepdo_wm_SelectFrame()
{
    // select base frame for current direction
    if (player.pshowdir==RIGHT) player.playframe = PMAPRIGHTFRAME;
    else if (player.pshowdir==LEFT) player.playframe = PMAPLEFTFRAME;
    else if (player.pshowdir==UP) player.playframe = PMAPUPFRAME;
    else if (player.pshowdir==DOWN) player.playframe = PMAPDOWNFRAME;

    // episode 3 map frames start at 31, ep1&2 at 32
    if (levelcontrol.episode==3) player.playframe--;

    // no walking animation if we're fading
    if (fade.mode != NO_FADE) return;

    // add in walk frame if walking
    if (player.pwalking) player.playframe += player.pwalkframe;
}

// set blockedl and blockedr...is Keen up against a solid object?
void gamepdo_wm_setblockedlrud()
{
   player.blockedl = player.blockedr = 0;
   player.blockedu = player.blockedd = 0;

   // R
   if (wm_issolid((player.x>>CSF)+8, (player.y>>CSF)+1))
      { player.blockedr = 1; }
   else if (wm_issolid((player.x>>CSF)+8, (player.y>>CSF)+8))
      { player.blockedr = 1; }
   else if (wm_issolid((player.x>>CSF)+8, (player.y>>CSF)+13))
      { player.blockedr = 1; }

   // L
   if (wm_issolid((player.x>>CSF)+0, (player.y>>CSF)+1))
      { player.blockedl = 1; }
   else if (wm_issolid((player.x>>CSF)+0, (player.y>>CSF)+8))
      { player.blockedl = 1; }
   else if (wm_issolid((player.x>>CSF)+0, (player.y>>CSF)+13))
      { player.blockedl = 1; }

   // U
   if (wm_issolid((player.x>>CSF)+1, (player.y>>CSF)-1))
      { player.blockedu = 1; }
   else if (wm_issolid((player.x>>CSF)+4, (player.y>>CSF)-1))
      { player.blockedu = 1; }
   else if (wm_issolid((player.x>>CSF)+7, (player.y>>CSF)-1))
      { player.blockedu = 1; }

   // D
   if (wm_issolid((player.x>>CSF)+1, (player.y>>CSF)+14))
      { player.blockedd = 1; }
   else if (wm_issolid((player.x>>CSF)+4, (player.y>>CSF)+14))
      { player.blockedd = 1; }
   else if (wm_issolid((player.x>>CSF)+7, (player.y>>CSF)+14))
      { player.blockedd = 1; }

}

// select proper player direction
void gamepdo_wm_setdir()
{
int exception;
   player.dpadcount = 0;
   if (player.keytable[KLEFT]) { player.pdir = LEFT; player.dpadcount++; }
   if (player.keytable[KRIGHT]) { player.pdir = RIGHT; player.dpadcount++; }
   if (player.keytable[KUP]) { player.pdir = UP; player.dpadcount++; }
   if (player.keytable[KDOWN]) { player.pdir = DOWN; player.dpadcount++; }

   exception = 0;
   if (player.pshowdir==UP && !player.keytable[KUP]) exception = 1;
   if (player.pshowdir==DOWN && !player.keytable[KDOWN]) exception = 1;
   if (player.pshowdir==LEFT && !player.keytable[KLEFT]) exception = 1;
   if (player.pshowdir==RIGHT && !player.keytable[KRIGHT]) exception = 1;
   if (player.keytable[KUP] && player.keytable[KDOWN]) player.pshowdir = DOWN;
   if (player.keytable[KLEFT] && player.keytable[KRIGHT]) player.pshowdir = RIGHT;

   if (player.dpadcount==1 || player.dpadlastcount==0 || exception)
   {
     player.pshowdir = player.pdir;
   }
   player.dpadlastcount = player.dpadcount;
}

// handles inertia and friction for the Y direction
// (this is where the inertia is actually applied to playx)
void gamepdo_InertiaAndFriction_Y()
{
   if (player.hideplayer)
   {
     player.pinertia_y = 0;
     return;
   }

   // if we hit a solid object kill all inertia
   if (player.pinertia_y > 0 && player.blockedd)
   {
     player.pinertia_y = 0;
   }
   else if (player.pinertia_y < 0 && player.blockedu)
   {
     player.pinertia_y = 0;
   }

   // apply pinertia_y
   player.y += player.pinertia_y;

   // if we stopped walking (i.e. LRUD not held down) apply friction
    if (!player.keytable[KDOWN] && !player.keytable[KUP])
    {

      // and apply friction to pinertia_y
      if (player.pfriction_timer_y > PFRICTION_RATE_WM)
      {
        if (player.pinertia_y < 0) player.pinertia_y++;
        else if (player.pinertia_y > 0) player.pinertia_y--;
        player.pfriction_timer_y = 0;
      }
      else player.pfriction_timer_y++;
    }

}

void gamepdo_wm_AllowEnterLevel()
{
int lvl;
int x,y,i,o;
int destx, desty;
int telfrom, teldest, telsnap;

// 3859 <-> 3889
// 3911 -> 3955
// 3877 <-> 3922

// 4032 -> 3842
// 3842 -> 3877
// 4063 -> 4094
// 4094 -> 4072
// 4072 -> 3980
// 3980 -> 4032 nosnap

// table of teleporters (mostly for ep3). player will teleport from
// a teleporter in a teleport_from entry, to the teleporter
// in the matching teleport_dest entry. Thus 3859 leads to 3889, etc.
// if the snap entry is 1 the screen will snap to center on the new
// position, otherwise it will smoothly scroll over to it
#define NUM_TPORTS    20
int teleport_from[NUM_TPORTS+1] = {4014,4032,3842,4063,4094,4072,3980,3859,3889,3911,3877,3922,3947,4025,3988,3955,41,38,0};
int teleport_dest[NUM_TPORTS+1] = {4072,3842,3877,4094,4072,3980,4032,3889,3859,3955,3922,3877,4025,3988,3911,3889,38,41,0};
int teleport_snap[NUM_TPORTS+1] = {1   ,1   ,1   ,1   ,1   ,1   ,0   ,0   ,0   ,1   ,1   ,1   ,1   ,0   ,1   ,1   ,0 ,0 ,0};

    if (fade.mode != NO_FADE) return;
    if (player.hideplayer) return;  // don't execute function while teleporting

    if ((player.keytable[KJUMP] || player.keytable[KPOGO]) && !player.wm_lastenterstate)
    {   // trying to enter a level (or use a teleporter, etc)

        // get level/object marker beneath player
        x = (player.x>>CSF)+4;
        y = (player.y>>CSF)+4;
        lvl = map.objectlayer[x>>4][y>>4];
        if (!lvl)
        {
          y = (player.y>>CSF)+8;
          lvl = map.objectlayer[x>>4][y>>4];
        }

        x = x >> 4 << 4;
        y = y >> 4 << 4;

        if (lvl)
        {
          /* handle special map objects, and standard levels too */

          // look through the teleporter tables and see if this is
          // a teleporter
          telfrom = 0;
          for(i=0;i<NUM_TPORTS;i++)
          {
            if (teleport_from[i]==lvl)
            { // we found it! it IS a teleporter!
              telfrom = teleport_from[i];
              teldest = teleport_dest[i];
              telsnap = teleport_snap[i];
              break;
            }
            else if (teleport_from[i]==0)
            { // nope. it's not a teleporter.
              break;
            }
          }

          // need to teleport somewhere?
          if (telfrom)
          {
             if (map_findobject(teldest, &destx, &desty))
             {
                o = spawn_object(x<<CSF,y<<CSF,OBJ_TELEPORTER);
                objects[o].ai.teleport.direction = TELEPORTING_OUT;
                objects[o].ai.teleport.destx = destx << 4 << CSF;
                objects[o].ai.teleport.desty = desty << 4 << CSF;
                objects[o].ai.teleport.NoExitingTeleporter = 0;
                objects[o].ai.teleport.snap = telsnap;
                if (levelcontrol.episode==3)
                {
                  objects[o].ai.teleport.baseframe = TELEPORT_BASEFRAME_EP3;
                  objects[o].ai.teleport.idleframe = TELEPORT_IDLEFRAME_EP3;
                }
                else 
                { // the teleporter in ep1
                  objects[o].ai.teleport.baseframe = TELEPORT_GRAY_BASEFRAME_EP1;
                  objects[o].ai.teleport.idleframe = TELEPORT_GRAY_IDLEFRAME_EP1;
                }
                sound_play(SOUND_TELEPORT, PLAY_NOW);

                player.hideplayer = 1;
             }            
          }
          else
          {
              switch(lvl)
              {
              case NESSIE_PATH: break;
              case NESSIE_PAUSE: break;
              case NESSIE_MOUNTPOINT: break;

              // ep1 bonus teleporter
              case LVLS_TELEPORTER_BONUS:
                o = spawn_object(x<<CSF,y<<CSF,OBJ_TELEPORTER);
                objects[o].ai.teleport.direction = TELEPORTING_OUT;
                objects[o].ai.teleport.destx = TELEPORT_BONUS_DESTX;
                objects[o].ai.teleport.desty = TELEPORT_BONUS_DESTY;
                objects[o].ai.teleport.baseframe = TELEPORT_RED_BASEFRAME_EP1;
                objects[o].ai.teleport.idleframe = TELEPORT_RED_IDLEFRAME_EP1;
                objects[o].ai.teleport.NoExitingTeleporter = 1;
                objects[o].ai.teleport.snap = 1;
                player.hideplayer = 1;
                break;

              case LVLS_SHIP:
                if (levelcontrol.episode==1)
                {
                  YourShipNeedsTheseParts();
                }
                else
                {
                  ShipEp3();
                }
                break;

              default:      // a regular level
                levelcontrol.chglevelto = (lvl & 0x7fff);
                endlevel(1);
                sound_play(SOUND_ENTER_LEVEL, PLAY_NOW);
                break;

              } // end switch(level)
          } // end isn't a teleporter
        } // end if(lvl)
    } // end trying to enter a level

    player.wm_lastenterstate = (player.keytable[KJUMP] || player.keytable[KPOGO]);
}

char wm_issolid(int xb, int yb)
{
  // for map tiles solidl and solidr are always gonna be the same...
  // so we can get away with this.
  if (tiles[getmaptileat(xb, yb)].solidl) return 1;
  if (getlevelat(xb, yb) & 0x8000) return 1;
  return 0;
}

// if nessie is at a mount point near the player, mounts him!
void MountNessieIfAvailable()
{
char AtSameMountPoint;

   if (!player.mounted)
   {
      // is nessie paused?
      if (objects[NessieObjectHandle].ai.nessie.pausetimer)
      {
         // is she at the same mount point the player is?
         if (player.y>>CSF>>4 < map.ysize>>2 && objects[NessieObjectHandle].y>>CSF>>4 < map.ysize>>2) AtSameMountPoint = 1;    // both at mortimer's castle
         else if (player.y>>CSF>>4 > map.ysize>>1 && objects[NessieObjectHandle].y>>CSF>>4 > map.ysize>>1) AtSameMountPoint = 1;    // both at secret island
         else AtSameMountPoint = 0;

         if (AtSameMountPoint)
         {
            objects[NessieObjectHandle].ai.nessie.mounted = 1;
            player.mounted = 1;
            player.hideplayer = 1;
         }
      }
   }
}

void AllowMountUnmountNessie()
{
int objmarker;
   if (!player.mounted)
   {  // not mounted. find out if he's trying to mount
      // if the upper quarter of the map (mortimer's castle mount point)
      // he's trying to mount if he's on a NESSIE_MOUNTPOINT object marker
      // and he's going right and is blockedr and/or is going down and is
      // blockedd. in the bottom quarter (secret island mount point)
      // it's up and blockedu.

      // make sure he's on a mount point
      objmarker = getlevelat((player.x>>CSF)+4, (player.y>>CSF)+9);
      if (objmarker != NESSIE_MOUNTPOINT) return;

      // is he trying to mount?
      if (player.y>>CSF>>4 < map.ysize>>2)
      {  // at mortimer's castle mount point
          if ((player.keytable[KRIGHT] && player.blockedr) || \
              (player.keytable[KDOWN] && player.blockedd))
          {
            // YES! if nessie is at that mount point, mount her!!
              MountNessieIfAvailable();
          }
      }
      else
      {  // at secret island mount point
          if (player.keytable[KUP] && player.blockedu)
          {
             MountNessieIfAvailable();
          }
      }
   }
   else
   {  // mounted. find out if he's trying to unmount.
      if (objects[NessieObjectHandle].ai.nessie.pausetimer)
      {  // nessie is paused
         if (objects[NessieObjectHandle].y>>CSF>>4 < map.ysize>>2)
         {  // nessie is at mortimer's castle mount point
            if (player.keytable[KUP])
            {
               // unmount nessie
               objects[NessieObjectHandle].ai.nessie.mounted = 0;
               player.mounted = 0;
               player.hideplayer = 0;
            }
         }
         else if (objects[NessieObjectHandle].y>>CSF>>4 > map.ysize>>1)
         {  // nessie is at secret island mount point
            if (player.keytable[KDOWN])
            {
               // unmount nessie
               objects[NessieObjectHandle].ai.nessie.mounted = 0;
               player.mounted = 0;
               player.hideplayer = 0;
               player.y += (18<<CSF);
               player.x += (8<<CSF);
            }
         }
      }

   }
}
