/* KEYDRV.C
  The low-level keyboard driver for the SDL port.
  Functions a keyboard driver must provide are KeyDrv_Start() and
  KeyDrv_Stop(). Once KeyDrv_Start() is called the keyboard driver
  needs to somehow (platform-specific) keep immediate_keytable[]
  filled up with the current key states.
*/

#include "keen.h"
#include "keenext.h"
#include <nds.h>

char KeyDrv_Start(void)
{
  VidDrv_printf("KeyDrv_Start()\n");
  return 0;
}

void KeyDrv_Stop(void)
{
}

// we process the cheat codes here so that it's not necessary to
// have a seperate keytable[] index for each cheat code letter
char key_g=0, key_o=0, key_d=0;

static int key_prev = 0;
static int key_curr = 0;

#define key_held(key) ((key_curr|key_prev) & key)
#define key_hit(key) ((key_curr&~key_prev) & key)

#define key_poll() { key_prev=key_curr; key_curr = KEYS_CUR; }
#define key_released(key) ((~key_curr&key_prev) & key)
#define pen_touch (((~IPC->buttons) >> 6) & 1)

/* poll_events() is called from SpeedThrottle() (timedrv.c) */
void poll_events(void)
{
  key_poll();
  immediate_keytable[KPAUSE] = (key_hit(KEY_SELECT)|| key_held(KEY_SELECT) ) ? 1 : 0;
  immediate_keytable[KSELECT] =  (key_hit(KEY_START)|| key_held(KEY_START) ) ? 1 : 0;
  immediate_keytable[KQUIT] = (key_hit(KEY_L)|| key_held(KEY_L) || SCR_TOUCH)  ? 1 : 0;
  immediate_keytable[KSAVE] = (key_hit(KEY_R)|| key_held(KEY_R) )  ? 1 : 0;
  
  immediate_keytable[KUP]     = (key_hit(KEY_UP)|| key_held(KEY_UP) ) ? 1 : 0;
  immediate_keytable[KRIGHT]  = (key_hit(KEY_RIGHT)|| key_held(KEY_RIGHT) ) ? 1 : 0;
  immediate_keytable[KDOWN]   = (key_hit(KEY_DOWN)|| key_held(KEY_DOWN) ) ? 1 : 0;
  immediate_keytable[KLEFT]   = (key_hit(KEY_LEFT)|| key_held(KEY_LEFT) ) ? 1 : 0;

  if (key_hit(KEY_A)|| key_held(KEY_A) ) { immediate_keytable[KNUM1] = 1; immediate_keytable[KOBF] = 1; }
  else { immediate_keytable[KNUM1] = 0; immediate_keytable[KOBF] = 0; }
  if (key_hit(KEY_B)|| key_held(KEY_B) ) { immediate_keytable[KNUM2] = 1; immediate_keytable[KJUMP] = 1; }
  else { immediate_keytable[KNUM2] = 0; immediate_keytable[KJUMP] = 0; }
  if (key_hit(KEY_X)|| key_held(KEY_X) ) { immediate_keytable[KNUM3] = 1; immediate_keytable[KN] = 1; immediate_keytable[KQ] = 1; }
  else { immediate_keytable[KNUM3] = 0; immediate_keytable[KN] = 0; immediate_keytable[KQ] = 0; }
  if (key_hit(KEY_Y)|| key_held(KEY_Y) ) { immediate_keytable[KNUM4] = 1; immediate_keytable[KPOGO] = 1; immediate_keytable[KT] = 1; immediate_keytable[KY] = 1; }
  else { immediate_keytable[KNUM4] = 0; immediate_keytable[KPOGO] = 0; immediate_keytable[KT] = 0; immediate_keytable[KY] = 0; }
}
