/* MAIN.C
  This is CloneKeen's main source file.

  The CloneKeen source may be freely distributed and
  modified as per the GPL but please give credit to
  the original author, Caitlin Shaw.

  Please go ahead and port this game to other platforms.
  I would love to see it on Linux, Mac, or CE but I do
  not have the platforms/knowledge to port to these platforms.

  If you make any changes or improvements to the code that
  you feel merit inclusion in the source tree email them
  to me at rogueeve@mailshack.com or get my latest email
  from the SourceForge site.

  Thanks to ID Software for the "Commander Keen: Invasion of
  the Vorticons" games. "Commander Keen" and it's associated
  graphics, level, and sound files are the property of ID
  Software. CloneKeen requires the original version of a
  Commander Keen game in order to be able to emulate that
  episode.

  CloneKeen comes with one new level for episode 1 (shareware),
  and 2 new levels for episode 2. This levels are released
  under the same license as the rest of the program.

  keen 17 -ep1 -fs
  keen 18 -ep2 -fs
  keen 19 -ep2 -fs

  Enjoy the Code
         -Caitlin

  CloneKeen 2003-2005 Caitlin Shaw
*/

#include "keen.h"
#include <unistd.h>
#include <nds.h>

#include "efs_lib.h"    // include EFS lib

extern int IntroCanceled;
int NessieObjectHandle;
int DemoObjectHandle;
int BlankSprite;
int DemoSprite;
int current_demo;
int framebyframe;
int fps=0, curfps=0;
unsigned int demo_RLERunLen;
unsigned char demo_data[DEMO_MAX_SIZE+1];
unsigned int demo_data_index;
int errorlog;

char QuitState = NO_QUIT;

stString strings[MAX_STRINGS+1];
int numStrings;

int demomode;
FILE *demofile;

char ScreenIsScrolling;
int gunfiretimer, gunfirefreq;

extern unsigned long scroll_x;
extern unsigned int scrollx_buf;
extern unsigned char scrollpix;
extern unsigned int mapx;
extern unsigned int mapxstripepos;
extern int scroll_y;
extern unsigned int scrolly_buf;
extern unsigned char scrollpixy;
extern unsigned int mapy;
extern unsigned int mapystripepos;

stLevelControl levelcontrol;

char loadinggame, loadslot;

FILE *errlog=NULL;

stFade fade;
stMap map;
unsigned int AnimTileInUse[ATILEINUSE_SIZEX][ATILEINUSE_SIZEY];
stTile tiles[MAX_TILES+1];
unsigned char tiledata[MAX_TILES+1][16][16];
stSprite sprites[MAX_SPRITES+1];
stBitmap bitmaps[MAX_BITMAPS+1];
stObject objects[MAX_OBJECTS+1];
char font[MAX_FONT+1][8][8];
stAnimTile animtiles[MAX_ANIMTILES+1];
stPlayer player;
stOption options[NUM_OPTIONS];
int episodeNum;
unsigned char scrollbuf[SCROLLBUF_MEMSIZE] __attribute((aligned(4)));
//unsigned char *scrollbuf;
//unsigned char *blitbuf = NULL;
char immediate_keytable[KEYTABLE_SIZE+1];
char last_immediate_keytable[KEYTABLE_SIZE+1];
int scroll_y = 0;
unsigned int objdefsprites[NUM_OBJ_TYPES+1];
char frameskiptimer=0;

int thisplayer;
char showfps;
char localmp;

int crashflag,crashflag2,crashflag3;
char *why_term_ptr = "No reason given.";

unsigned int efsdemomode;

void setoption(int opt, char *name, char value)
{
  if (name != NULL)
    options[opt].name = name;

  options[opt].value = value;
}

void SetDefaultOptions(void)
{
	setoption(OPT_FRAMESKIP, "Frameskip", 5);
	setoption(OPT_SNDVOL, "Sound Volume", 10);
	setoption(OPT_ONEBUTTONFIRE, "One-button Firing", 1);
	setoption(OPT_FULLYAUTOMATIC, "Fully Automatic Raygun", 0);
	setoption(OPT_SUPERPOGO, "SuperPogo", 0);
	setoption(OPT_MEAN, "Increase Difficulty", 0);
}

void on_irq() 
{	

}

int main(int argc, char **argv)
{
//  int i;
  int opt = 0;
  int dtm = 0;
  int retval;
  int eseq;
  int defaultopt;
  bool isFATSystem;

  //set the mode for 2 text layers and two extended background layers
	powerON(POWER_ALL_2D); 

  // Init irq
  irqInit();
	irqSet(IRQ_VBLANK, on_irq);
	irqEnable(IRQ_VBLANK);	
  
  // Little message during efs initialisation
	videoSetMode(MODE_5_2D | DISPLAY_BG3_ACTIVE); 
  videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
  vramSetMainBanks(VRAM_A_MAIN_BG, VRAM_B_MAIN_BG,  VRAM_C_SUB_BG , VRAM_D_SUB_SPRITE); 

  // Little message during efs initialisation
	videoSetMode(MODE_0_2D | DISPLAY_BG0_ACTIVE); 
  videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
  vramSetMainBanks(VRAM_A_MAIN_BG, VRAM_B_MAIN_BG,  VRAM_C_SUB_BG , VRAM_D_SUB_SPRITE); 
  SUB_BG0_CR = BG_MAP_BASE(31);
  BG_PALETTE_SUB[255] = RGB15(31,31,31);	
  consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);
  iprintf("Please wait ...\n");

  // init EFSlib & libfat
  isFATSystem = EFS_Init(EFS_AND_FAT,NULL);
  swiWaitForVBlank(); swiWaitForVBlank();

  if (!isFATSystem) { iprintf("EFS init failed :(...\n"); while(1); }
  EFS_ChDefaultDir();
  
  while(1) {  
    // Show the menu
    showMenuDS();
    
    // launch the game
    errorlog = 0;
    demomode = DEMO_NODEMO;
    current_demo = 1;
    framebyframe = 0;
    debugmode = 0;
    acceleratemode = 0;
    eseq = 0;
    localmp = 1;
    showfps = 0;

    VidDrv_InitConsole();

    banner();
    VidDrv_printf("\n");

    // set default config-menu options
    SetDefaultOptions();

    restart_game: ;
    episodeNum = levelcontrol.episode;
      
    demofile = NULL;

    // allocate memory for scrollbuffer, etc.
    if (allocmem()) {crashflag = 1; goto abort; };

    // decode graphics from EGALATCH & EGASPRIT
    if (latch_loadgraphics(levelcontrol.episode)) {crashflag = 1; goto abort; };
    if (loadstrings("efs:/strings.dat")) {crashflag = 1; goto abort; };

    VidDrv_printf("Starting SDL...\n");

    /* initilize/activate all drivers */
    VidDrv_printf("Starting sound driver...\n");
    if (SoundDrv_Start()) {crashflag = 1; goto abort; };
    VidDrv_printf("Starting keyboard driver...\n");
    if (KeyDrv_Start()) {crashflag = 1; goto abort; };
    VidDrv_printf("Starting graphics driver...\n");
    if (Graphics_Start()) {crashflag = 1; goto abort; };
    VidDrv_printf("Starting timer driver...\n");
    if (TimeDrv_Start()) {crashflag = 1; goto abort; };

    player.x = player.y = 0;
    initgamefirsttime();
    initgame();
    if (dtm) goto directtomap;

    VidDrv_printf("Game starting...\n");

    if (eseq)
    {
      endsequence();
      goto exitt;
    }

    if (intro()) goto exitt;

    defaultopt = 0;
    do   
    {
      if (QuitState==QUIT_TO_TITLE) QuitState = NO_QUIT;
      VidDrv_printf("QUITSTATE: %d\n", QuitState);
      VidDrv_printf("calling mainmenu()\n");
      opt = mainmenu(defaultopt);
      VidDrv_printf("gcl: opt = %d\n", opt);
      defaultopt = 0;
      IntroCanceled = 0;
      switch(opt)
      {
        case MAINMNU_1PLAYER:
directtomap: ;
          defaultopt = 0;
          current_demo = 1;
          initgamefirsttime();
          loadinggame = 0;
          playgame_levelmanager(argc, argv, dtm);
          break;
        case MAINMNU_LOADGAME:
          if (loadslot)
          {
            loadinggame = 1;
            defaultopt = 0;
            current_demo = 1;
            initgamefirsttime();
            playgame_levelmanager(argc, argv, dtm);
          }
          break;
        case MAINMNU_TIMEOUT:
        case MAINMNU_DEMO:
          retval = play_demo(current_demo);
          if (retval==DEMO_RESULT_FILE_BAD)
          {
            // we tried to play a demo that did not exist--assume we
            // reached the last demo and go back to the intro
            intro();
            current_demo = 0;
          }
          else if (retval==DEMO_RESULT_CANCELED)
          { // user hit a key to cancel demo
            IntroCanceled = 1;            // pop up menu
          }

          if (IntroCanceled)
          { // user canceled out of demo (or intro if at end of demos)
          // if user selected "demo" have it selected when he comes back
            if (opt==MAINMNU_DEMO)
            {
              defaultopt = MAINMNU_DEMO;
            }
          }
          current_demo++;
          break;
        case RESTART_GAME:
          VidDrv_printf("********************\n");
          VidDrv_printf(" Restarting game...\n");
          VidDrv_printf("********************\n\n");
          cleanup();
          goto restart_game;
          break;
        default: break;
      }
      VidDrv_printf("bottom of game control loop opt=%d crashflag=%d\n", opt, crashflag);
    } while(opt != MAINMNU_QUIT && !crashflag);
  
    opt = 0;
    dtm = 0;
    QuitState = NO_QUIT;

VidDrv_printf("WARNING! WARNING! dropped out of game control loop!!!!\n");
exitt: ;
    Graphics_Stop();
    banner();
    VidDrv_printf("\nThanks for playing!\n\n");
    cleanup();
    if (crashflag)
    {
#ifndef DEBUG
      showMenuDSError();
#endif      
      //if (crashflag != QUIT_NONFATAL) VidDrv_printf("\a");
      VidDrv_printf("abnormal program termination, error code %d/%d/%d\nexplanation: %s\n", crashflag,crashflag2,crashflag3, why_term_ptr);
      //if (crashflag != QUIT_NONFATAL) goto startup_error;
    }
  
//	return 0;

abort: ;
    if (crashflag) {
      VidDrv_printf("Fatal error, cleaning up.\n");
      cleanup();
    }
    goto startup_error;

startup_error: ;
    if (crashflag){
#ifndef DEBUG
      showMenuDSError();
#endif
      VidDrv_printf("A fatal error has occurred; game is shutting down.\n");
    }
  }

  return 0;
}


void playgame_levelmanager(int argc, char **argv, int dtm)
{
  int o, wm, firsttime = 1;
  char levelname[80];
  char SaveGameFileName[40];
  int newlevel;

  levelcontrol.command = LVLC_CHANGE_LEVEL;
  levelcontrol.chglevelto = WM_MAP_NUM;
  levelcontrol.tobonuslevel = 0;
  levelcontrol.success = 0;
  map.firsttime = 1;
  do
  {
    initgame();

    if (argc>1 && *argv[1] != '-' && dtm==0)
    {
      levelcontrol.chglevelto = atoi(argv[1]);
      if (!levelcontrol.chglevelto) levelcontrol.chglevelto = 1;
    }

    newlevel = levelcontrol.chglevelto;
    if (levelcontrol.episode==1 && options[OPT_MEAN].value)
    {
        // in high-difficulity mode switch levels 5 & 9 so
        // you can't get the pogo stick until you make it
        // to the dark side of mars.
        if (newlevel==5)
           newlevel = 9;
        else if (newlevel==9)
           newlevel = 5;
    }
    sprintf(levelname, "LEVEL%02d.CK%d", newlevel, levelcontrol.episode);

    if (levelcontrol.chglevelto==WORLD_MAP)
    {
      wm = 1;
    }
    else
    {
      wm = 0;
    }
    levelcontrol.canexit = 1;   // assume can exit before loading map
    if (loadmap(levelname, levelcontrol.chglevelto, wm))
    {
      crashflag = 1;
      crashflag2 = levelcontrol.chglevelto;
      why_term_ptr = "Unable to load the map (# shown in crashflag2).";
    }
    levelcontrol.curlevel = levelcontrol.chglevelto;
    if (levelcontrol.curlevel == FINAL_MAP)
    {
      levelcontrol.isfinallevel = 1;
      levelcontrol.canexit = 0;
    }
    else
    {
      levelcontrol.isfinallevel = 0;
    }

    if (firsttime)
    {
        player.mapplayx = player.x;
        player.mapplayy = player.y;
    }
    firsttime = 0;

    levelcontrol.command = LVLC_NOCOMMAND;

    levelcontrol.dark = 0;
    if (loadinggame)
    {
      sprintf(SaveGameFileName, "ep%csave%c.dat", levelcontrol.episode+'0', loadslot+'0');
      wm = savegameiswm(SaveGameFileName);
      if (game_load(SaveGameFileName))
      {
        crashflag = 1;
        crashflag2 = loadslot;
        why_term_ptr = "Error loading game (slot # in flag2)! The save file may be corrupt or created by a different version of CloneKeen.";
        return;
      }
    }
    pal_init(levelcontrol.dark);

    if (wm)
    {  // entering map from normal level, or first time around
      if (!levelcontrol.tobonuslevel)
      {
        if (!loadinggame)
        {
            player.x = player.mapplayx;
            player.y = player.mapplayy;
        }
      }
      else
      {  // respawn at the bonus level
        player.x = BONUSLEVEL_RESPAWN_X;
        player.y = BONUSLEVEL_RESPAWN_Y;
        if (player.isPlaying && player.inventory.lives)
        {
          player.hideplayer = 1;
          o = spawn_object((player.x>>CSF>>4)<<CSF<<4,((player.y>>CSF>>4)+1)<<CSF<<4,OBJ_TELEPORTER);
          objects[o].ai.teleport.direction = TELEPORTING_IN;
          //objects[o].ai.teleport.whichplayer = i;
          objects[o].ai.teleport.baseframe = TELEPORT_RED_BASEFRAME_EP1;
          objects[o].ai.teleport.idleframe = TELEPORT_RED_IDLEFRAME_EP1;
          sound_play(SOUND_TELEPORT, PLAY_NOW);
        }
      }

      if (!levelcontrol.success || firsttime)
      {
        if (!levelcontrol.tobonuslevel) levelcontrol.dokeensleft = 1; 
        // when you die you lose all keycards
        if (player.isPlaying)
        {
          if (levelcontrol.episode!=3)
          {
            take_keycard(DOOR_YELLOW);
            take_keycard(DOOR_RED);
            take_keycard(DOOR_GREEN);
            take_keycard(DOOR_BLUE);
          }
          else
          {
            take_keycard(DOOR_YELLOW_EP3);
            take_keycard(DOOR_RED_EP3);
            take_keycard(DOOR_GREEN_EP3);
            take_keycard(DOOR_BLUE_EP3);
          }
        }
	  }
      else levelcontrol.dokeensleft = 0;

      gameloop();

      player.mapplayx = player.x;
      player.mapplayy = player.y;
    }
    else
    {  // entering a normal level from map
       levelcontrol.dokeensleft = 0;
       gameloop();

       // after completion of a normal level check if the game is won
       if (gameiswon())
       {
          levelcontrol.command = LVLC_END_SEQUENCE;
       }
    }
  } while(levelcontrol.command==LVLC_CHANGE_LEVEL && !crashflag);

  if (levelcontrol.command==LVLC_END_SEQUENCE && !levelcontrol.gameovermode)
  {
    endsequence();
    VidDrv_printf("eseq complete\n");
  }
  else if (levelcontrol.command==LVLC_TANTALUS_RAY)
  {
    eseq2_vibrate();
    eseq2_TantalusRay();
    IntroCanceled = 1;               // popup main menu immediately
  }
}

// plays the demo file specified in fname
// returns:
//  DEMO_RESULT_FILE_BAD               demo does not exist or file corrupt
//  DEMO_RESULT_COMPLETED              demo played all the way through
//  DEMO_RESULT_CANCELED               user canceled the demo
char play_demo(int demonum)
{
int i;
int byt;
int lvl;
char filename[40];
char SaveOptions[NUM_OPTIONS];

   /* open the demo file */
   sprintf(filename, "efs:/ep%ddemo%d.dat", levelcontrol.episode, demonum);
   demofile = fopen(filename, "rb");
   if (!demofile)
   {    
     return DEMO_RESULT_FILE_BAD;
   }

   /* read in the header */
   if (fgetc(demofile) != 'D') goto demoHeaderCorrupt;
   if (fgetc(demofile) != 'M') goto demoHeaderCorrupt;
   if (fgetc(demofile) != 'O') goto demoHeaderCorrupt;
   if (fgetc(demofile) != levelcontrol.episode) goto demoHeaderCorrupt;
   lvl = fgetc(demofile);

   /* load the compressed demo into the demo_data[] array */
   for(i=0;i<DEMO_MAX_SIZE;i++)
   {
     byt = fgetc(demofile);     // get byte from file
     if (byt<0) goto gotEOF;    // check for EOF
     demo_data[i] = byt;        // copy to demo_data[]
   }
   why_term_ptr = "Demo file is too big! (Increase DEMO_MAX_SIZE)";
   crashflag = 1;
gotEOF: ;
   fclose(demofile);
   demofile = NULL;

   /* initilize some variables */
   demo_RLERunLen = 0;
   demo_data_index = 0;
   demomode = DEMO_PLAYBACK;
   loadinggame = 0;
   levelcontrol.curlevel = lvl;
   levelcontrol.command = LVLC_NOCOMMAND;

   initgamefirsttime();
   initgame();

   /* now load the map and play the level */
   sprintf(filename, "level%02d.ck%d", levelcontrol.curlevel, levelcontrol.episode);
   if (loadmap(filename, levelcontrol.curlevel, 0)) return DEMO_RESULT_FILE_BAD;

   for(i=0;i<NUM_OPTIONS;i++) SaveOptions[i] = options[i].value;
   SetDefaultOptions();
   gameloop();
   for(i=0;i<NUM_OPTIONS;i++) options[i].value = SaveOptions[i];

   /* based on success/failure returned from gameloop let the
      calling procedure know whether the user canceled or not */

   if (levelcontrol.success)
   {
     return DEMO_RESULT_COMPLETED;
   }
   else
   {
     return DEMO_RESULT_CANCELED;
   }

// this label is jumped to when there's an error reading the header.
// it closes the demo file and aborts.
demoHeaderCorrupt: ;
   fclose(demofile);
   demofile = NULL;
   return DEMO_RESULT_FILE_BAD;
}
