/* VIDDRV.C
  This file contains low-level graphics functions which are
  platform-specific.
*/
#include "keen.h"
#include "keenext.h"
#include <stdarg.h>

#include <nds.h>

char blitsurface_alloc = 0;
void VidDrv_printf(const char *str, ...);

#define WINDOW_STD_WIDTH            320
#define WINDOW_STD_HEIGHT           200
int window_width;
int window_height;

// pointer to the line in VRAM to start blitting to when stretchblitting.
// this may not be the first line on the display as it is adjusted to
// center the image on the screen when in fullscreen.
unsigned char *VRAMPtr;

//SDL_Rect dstrect;

extern FILE *errlog;
#define CKLOGFILENAME            "ck.log"

#define MAX_CONSOLE_MESSAGES     3
#define CONSOLE_MESSAGE_X        3
#define CONSOLE_MESSAGE_Y        3
#define CONSOLE_MESSAGE_SPACING  9
#define CONSOLE_EXPIRE_RATE      250
typedef struct stConsoleMessage
{
  char msg[80];
} stConsoleMessage;
stConsoleMessage cmsg[MAX_CONSOLE_MESSAGES];
int NumConsoleMessages = 0;
int ConsoleExpireTimer = 0;

// "Console" here refers to the capability to pop up in-game messages
// in the upper-left corner during game play ala Doom.
void DrawConsoleMessages(void)
{
int i;
int y;

 if (!NumConsoleMessages) return;
 if (!ConsoleExpireTimer)
 {
   NumConsoleMessages--;
   if (!NumConsoleMessages) return;
   ConsoleExpireTimer = CONSOLE_EXPIRE_RATE;
 }
 else ConsoleExpireTimer--;

 y = CONSOLE_MESSAGE_Y;
 for(i=0;i<NumConsoleMessages;i++)
 {
    font_draw(cmsg[i].msg, CONSOLE_MESSAGE_X, y, 1);
    y += CONSOLE_MESSAGE_SPACING;
 }
}

// removes all console messages
void DeleteConsoleMsgs(void)
{
  NumConsoleMessages = 0;
}

// adds a console msg to the top of the screen and scrolls any
// other existing messages downwards
void AddConsoleMsg(char *the_msg)
{
int i;
  for(i=MAX_CONSOLE_MESSAGES-2;i>=0;i--)
  {
    strcpy(cmsg[i+1].msg, cmsg[i].msg);
  }
  strcpy(cmsg[0].msg, the_msg);

  if (NumConsoleMessages < MAX_CONSOLE_MESSAGES) NumConsoleMessages++;
  ConsoleExpireTimer = CONSOLE_EXPIRE_RATE;
}

void sb_blit(void)
{
  swiWaitForVBlank();

  DrawConsoleMessages();

  BG3_CX = (scrollx_buf<<8); 
  BG3_CY = (scrolly_buf<<8); 
  dmaCopy(scrollbuf,BG_GFX, 512*512);
}

void update_screen(void)
{
  swiWaitForVBlank();
}

// functions to directly set and retrieve pixels from the VGA display
void setpixel(int x, int y, unsigned char c)
{
  x = (scrollx_buf+x); if (x>511) x -= 511;
  y = (scrolly_buf+y); if (y>511) y -= 511;
  unsigned short *pixels = (unsigned short *) (0x06000000);
  pixels += (y<<8) + (x>>1);
  unsigned short value = (x & 1 ? ((c<<8) | (*(pixels) & 0xFF)) : (c | (*(pixels) & 0xFF00) ));
  *pixels = value;
}

unsigned char getpixel(int x, int y)
{
  return 0;
}

typedef struct {
    int r;
    int g;
    int b;
} plt;

static plt MyPalette[256*3];

// alter the color palette. the palette is not actually altered
// on-screen until pal_apply() is called.
void pal_set(int colour, int red, int green, int blue)
{
  MyPalette[colour].r = red;
  MyPalette[colour].g = green;
  MyPalette[colour].b = blue;
}

// applies all changes to the palette made with pal_set
void pal_apply(void)
{
  unsigned short i;

  for (i = 0; i < 256; i++) {
    BG_PALETTE[i]=RGB15(MyPalette[i].r>>3,MyPalette[i].g>>3,MyPalette[i].b>>3);

  }
}

void vblankIntr() {
  static const u16 jitter2[] = {
    0x40, 0xc0,		/* 0.25, 0.75 */
    0xc0, 0x40,		/* 0.75, 0.25 */
  };

  BG3_XDX = ((WINDOW_STD_WIDTH / 256) << 8) | (WINDOW_STD_WIDTH % 256);
  BG3_XDY = 0;
  BG3_YDX = 0;
  BG3_YDY = ((WINDOW_STD_HEIGHT / 192) << 8) | ((WINDOW_STD_HEIGHT % 192) ) ; 
  BG2_XDX = ((WINDOW_STD_WIDTH / 256) << 8) | (WINDOW_STD_WIDTH % 256);
  BG2_XDY = 0;
  BG2_YDX = 0;
  BG2_YDY = ((WINDOW_STD_HEIGHT / 192) << 8) | ((WINDOW_STD_HEIGHT % 192) ) ; 

  BG3_CX = (scrollx_buf<<8)+jitter2[0]; 
  BG3_CY = (scrolly_buf<<8)+jitter2[1]; 
  BG2_CX = (scrollx_buf<<8)+jitter2[2]; 
  BG2_CY = (scrolly_buf<<8)+jitter2[3]; 
}
  
// starts up the video driver
char VidDrv_Start(void)
{
  videoSetMode(MODE_5_2D | DISPLAY_BG2_ACTIVE | DISPLAY_BG3_ACTIVE); 
  vramSetBankA(VRAM_A_MAIN_BG_0x06000000);
  vramSetBankB(VRAM_B_MAIN_BG_0x06020000);

	BG2_CR = BG_BMP8_512x512 | BG_WRAP_ON;
	BG3_CR = BG_BMP8_512x512 | BG_WRAP_ON;
  
  BG3_XDX = ((WINDOW_STD_WIDTH / 256) << 8) | (WINDOW_STD_WIDTH % 256);
  BG3_XDY = 0;
  BG3_YDX = 0;
  BG3_YDY = ((WINDOW_STD_HEIGHT / 192) << 8) | ((WINDOW_STD_HEIGHT % 192) ) ; 
  BG3_CX = 0;
  BG3_CY = 0; 

  BG2_XDX = ((WINDOW_STD_WIDTH / 256) << 8) | (WINDOW_STD_WIDTH % 256);
  BG2_XDY = 0;
  BG2_YDX = 0;
  BG2_YDY = ((WINDOW_STD_HEIGHT / 192) << 8) | ((WINDOW_STD_HEIGHT % 192) ) ; 
  BG2_CX = 0;
  BG2_CY = 0; 

	irqSet(IRQ_VBLANK, vblankIntr);
  
  BLEND_CR = BLEND_ALPHA | BLEND_SRC_BG2 | BLEND_DST_BG3;
  BLEND_AB = (8 << 8) | 8; /* 50% / 50% */
  
/*
  if (!scrollbuf)
  {
    VidDrv_printf("VidDrv_Start: Can not initilize SDL graphics until scroll buffer is allocated (allocmem())!\n");
    return 1;
  }
  
  SDL_WM_SetCaption("CloneKeen", NULL);
  
  // When the program is through executing, call SDL_Quit 
*/
  return VidDrv_CreateSurfaces();
}

char VidDrv_CreateSurfaces(void)
{
  return 0;
}

// shuts down the video driver
void VidDrv_Stop(void)
{
}

// resets graphics to allow changing of resolution or zoom settings
void VidDrv_reset(void)
{
     VidDrv_printf("restart: shutting down viddrv\n");
     VidDrv_Stop();

     // re-allocate memory as blitbuffer may appear/disappear
     freemem();
     if (allocmem()) { crashflag = 1; return; }

     VidDrv_printf("restart: restarting viddrv\n");
     VidDrv_CreateSurfaces();   // don't re-call sdl_init*/
     pal_fade(PAL_FADE_SHADES);
}

// initilize the console that VidDrv_printf outputs to. For win32
// our "console" is just a log file, so here it just clears the log.
// With futures ports this function should be expanded, as this
// file should be platform indepedent (it is supposed to be
// SDL-specific, not win32-specific)
void VidDrv_InitConsole(void)
{
#ifdef DEBUG
  videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
  vramSetBankC(VRAM_C_SUB_BG);
  vramSetBankD(VRAM_D_SUB_SPRITE);
  SUB_BG0_CR = BG_MAP_BASE(31);
  BG_PALETTE_SUB[255] = RGB15(31,31,31);	
  consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);
#endif
}

void VidDrv_printf(const char *str, ...)
{
#ifdef DEBUG
  va_list argptr;
  char s[1024];

  /* prepare message */
  va_start(argptr, str);
  vsprintf(s, str, argptr);
  va_end(argptr);
  printf(s);
#endif
}
